import logging
from abc import ABC, abstractmethod
from typing import List, Optional

from ..state import AnalysisState, StateCheckMixin

logger = logging.getLogger(__name__)


class AbstractVisualization(ABC, StateCheckMixin):
    """
    Base class for visualization functionality.

    Provides basic functionality for namespace management and helper method to access frequently-used methods.

    Specifying namespace would narrow visibility scope to specific subspace of `state`. Namespaces
    can be specified in a nested form: `ns_a.ns_b.ns_c`. Please see :py:class:`~autogluon.eda.analysis.base.Namespace`
    wrapper on how to create namespaces.

    The main entry method of analysis is `render` function. When called, the execution flow is the following:
    - narrow `state` scope to specified `namespace`
    - call `_render` function for each component that returned `True` from `can_handle` call

    Parameters
    ----------
    namespace: str
        namespace to use; can be nested like `ns_a.ns_b.ns_c`
    kwargs

    See Also
    --------
    :py:class:`~autogluon.eda.analysis.base.Namespace`
    """

    def __init__(self, namespace: Optional[str] = None, **kwargs) -> None:
        super().__init__()
        self.namespace: List[str] = []
        self._kwargs = kwargs
        if namespace is not None:
            self.namespace = namespace.split(".")

    def _get_namespace_state(self, state):
        if self.namespace is not None:
            for k in self.namespace:
                state = state[k]
        return state

    @abstractmethod
    def can_handle(self, state: AnalysisState) -> bool:
        """
        Checks if state has all the required parameters for visualization.
        See also :func:`at_least_one_key_must_be_present` and :func:`all_keys_must_be_present` helpers
        to construct more complex logic.

        Parameters
        ----------
        state: AnalysisState
            fitted state

        Returns
        -------
            `True` if all the pre-requisites for rendering are present

        """
        raise NotImplementedError

    @abstractmethod
    def _render(self, state: AnalysisState) -> None:
        """
        @override
        Component-specific rendering logic.
        This method is designed to be overridden by the component developer.

        Parameters
        ----------
        state

        Returns
        -------

        """
        raise NotImplementedError

    def render(self, state: AnalysisState) -> None:
        """
        Render component.

        Parameters
        ----------
        state: AnalysisState
            state to render

        """
        state = self._get_namespace_state(state)
        if self.can_handle(state):
            self._render(state)



from typing import Callable, Dict, List, Optional, Union

from IPython.display import display
from ipywidgets import HBox, Layout, Output, Tab

from .. import AnalysisState
from .base import AbstractVisualization
from .jupyter import JupyterMixin

__all__ = [
    "MarkdownSectionComponent",
    "SimpleVerticalLinearLayout",
    "SimpleHorizontalLayout",
    "PropertyRendererComponent",
    "TabLayout",
]

from ..state import is_key_present_in_state


class SimpleVerticalLinearLayout(AbstractVisualization):
    """
    Renders facets in a sequential order (facets will appear in a vertical layout).
    """

    def __init__(
        self,
        facets: Union[AbstractVisualization, List[AbstractVisualization]],
        namespace: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(namespace, **kwargs)
        if not isinstance(facets, list):
            facets = [facets]
        self.facets = facets
        self._kwargs = kwargs

    def can_handle(self, state: AnalysisState) -> bool:
        return True

    def _render(self, state: AnalysisState) -> None:
        for facet in self.facets:
            facet.render(state)


class SimpleHorizontalLayout(SimpleVerticalLinearLayout):
    """
    Render components horizontally using `HBox` widget.
    See `HBox widget <https://ipywidgets.readthedocs.io/en/stable/examples/Widget%20List.html#HBox>`_ documentation for details.
    """

    def _render(self, state: AnalysisState) -> None:
        outs = [Output() for _ in range(len(self.facets))]
        for out, facet in zip(outs, self.facets):
            with out:
                facet.render(state)

        display(HBox(outs, layout=Layout(flex="row wrap")))


class TabLayout(SimpleVerticalLinearLayout):
    """
    Render components using `Tab` widget.
    See `HBox widget <https://ipywidgets.readthedocs.io/en/stable/examples/Widget%20List.html#Tabs>`_ documentation for details.

    """

    def __init__(self, facets: Dict[str, AbstractVisualization], namespace: Optional[str] = None, **kwargs) -> None:
        self.facet_tab_names = list(facets.keys())
        super().__init__(list(facets.values()), namespace, **kwargs)

    def _render(self, state: AnalysisState) -> None:
        outs = [Output() for _ in self.facets]
        tab = Tab(children=outs, width=400)
        for i, name in enumerate(self.facet_tab_names):
            tab.set_title(i, name)
        display(tab)
        for out, facet in zip(outs, self.facets):
            with out:
                facet.render(state)


class MarkdownSectionComponent(AbstractVisualization, JupyterMixin):
    """
    Render provided string as a Markdown cell.
    See `Jupyter Markdown cell <https://jupyter-notebook.readthedocs.io/en/stable/examples/Notebook/Working%20With%20Markdown%20Cells.html>`_
    documentation for details.

    Parameters
    ----------
    markdown: str
        markdown text to render
    condition_fn: Optional[Callable], default = None
        if specified, call the provided function with `state` arg. The function expected to return bool value
    namespace: Optional[str], default = None
        namespace to use; can be nested like `ns_a.ns_b.ns_c`

    """

    def __init__(
        self, markdown: str, condition_fn: Optional[Callable] = None, namespace: Optional[str] = None, **kwargs
    ) -> None:
        """

        Parameters
        ----------
        markdown
        condition_fn
        namespace
        kwargs
        """
        super().__init__(namespace, **kwargs)
        self.markdown = markdown
        self.condition_fn = condition_fn

    def can_handle(self, state: AnalysisState) -> bool:
        return True if self.condition_fn is None else self.condition_fn(state)

    def _render(self, state: AnalysisState) -> None:
        self.render_markdown(self.markdown)


class PropertyRendererComponent(AbstractVisualization, JupyterMixin):
    """
    Render component stored in `state`'s dot-separated path to property (i.e. `a.b` results in `state.a.b`)

    Parameters
    ----------
    property: str
        dot-separated path to property (i.e. `a.b` results in `state.a.b`)
    transform_fn: Optional[Callable], default = None
        if specified, call the provided function with the object extracted from `state`'s `property`.
        Returned transformation is then passed into render function
    namespace: Optional[str], default = None
        namespace to use; can be nested like `ns_a.ns_b.ns_c`
    """

    def __init__(
        self, property: str, transform_fn: Optional[Callable] = None, namespace: Optional[str] = None, **kwargs
    ) -> None:
        super().__init__(namespace, **kwargs)
        self.property = property
        self.transform_fn = transform_fn

    def can_handle(self, state: AnalysisState) -> bool:
        return is_key_present_in_state(state, self.property)

    def _render(self, state: AnalysisState) -> None:
        obj = state
        for p in self.property.split("."):
            obj = obj[p]
        if self.transform_fn is not None:
            obj = self.transform_fn(obj)
        self.display_obj(obj)



import logging
from typing import Optional

import matplotlib.pyplot as plt
import missingno as msno

from .. import AnalysisState
from .base import AbstractVisualization
from .jupyter import JupyterMixin

__all__ = ["MissingValues"]
logger = logging.getLogger(__name__)


class MissingValues(AbstractVisualization, JupyterMixin):
    """
    Renders visualization of missingness for datasets using one of the methods specified in `graph_type'.

    This visualization depends on :py:class:`~autogluon.eda.analysis.missing.MissingValuesAnalysis` analysis.

    See also `missingno <https://github.com/ResidentMario/missingno>`_ documentation

    Parameters
    ----------
    graph_type: str, default = 'matrix'
        One of the following visualization types:
        - matrix - nullity matrix is a data-dense display which lets you quickly visually pick out patterns in data completion
            This visualization will comfortably accommodate up to 50 labelled variables.
            Past that range labels begin to overlap or become unreadable, and by default large displays omit them.
        - bar - visualizes how many rows are non-null vs null in the column. Logarithmic scale can by specifying `log=True` in `kwargs`
        - heatmap - correlation heatmap measures nullity correlation: how strongly the presence or absence of one
            variable affects the presence of another. Nullity correlation ranges from -1
            (if one variable appears the other definitely does not) to 0 (variables appearing or not appearing have no effect on one another)
            to 1 (if one variable appears the other definitely also does).
            Entries marked <1 or >-1 have a correlation that is close to being exactingly negative or positive but is still not quite perfectly so.
        - dendrogram - the dendrogram allows to more fully correlate variable completion, revealing trends deeper than the pairwise ones
            visible in the correlation heatmap. The dendrogram uses a hierarchical clustering algorithm (courtesy of scipy) to bin variables
            against one another by their nullity correlation (measured in terms of binary distance).
            At each step of the tree the variables are split up based on which combination minimizes the distance of the remaining clusters.
            The more monotone the set of variables, the closer their total distance is to zero, and the closer their average distance (the y-axis) is to zero.
    headers: bool, default = False
        if `True` then render headers
    namespace: str, default = None
        namespace to use; can be nested like `ns_a.ns_b.ns_c`

    See Also
    --------
    :py:class:`~autogluon.eda.analysis.missing.MissingValuesAnalysis`
    """

    __OPERATIONS_MAPPING = {
        "matrix": msno.matrix,
        "bar": msno.bar,
        "heatmap": msno.heatmap,
        "dendrogram": msno.dendrogram,
    }

    MAX_MATRIX_VARIABLES_NUMBER = 50

    def __init__(
        self, graph_type: str = "matrix", headers: bool = False, namespace: Optional[str] = None, **kwargs
    ) -> None:
        super().__init__(namespace, **kwargs)
        self.graph_type = graph_type
        assert (
            self.graph_type in self.__OPERATIONS_MAPPING
        ), f"{self.graph_type} must be one of {self.__OPERATIONS_MAPPING.keys()}"
        self.headers = headers

    def can_handle(self, state: AnalysisState) -> bool:
        can_handle = self.all_keys_must_be_present(state, "missing_statistics")
        if can_handle and self.graph_type == "matrix" and self._has_too_many_variables_for_matrix(state):
            logging.warning(
                f"The dataset has more than {self.MAX_MATRIX_VARIABLES_NUMBER} variables; "
                f"matrix visualization will comfortably accommodate up to {self.MAX_MATRIX_VARIABLES_NUMBER} labelled variables. "
                f"Past that range labels begin to overlap or become unreadable, and by default large displays omit them."
            )
        return can_handle

    def _render(self, state: AnalysisState) -> None:
        for ds, ds_state in state.missing_statistics.items():
            self.render_header_if_needed(state, f"`{ds}` missing values analysis", ds=ds)
            widget = self._get_operation(self.graph_type)
            self._internal_render(widget, ds_state.data, **self._kwargs)

    def _has_too_many_variables_for_matrix(self, state: AnalysisState):
        for _, ds_state in state.missing_statistics.items():
            if len(ds_state.data.columns) > self.MAX_MATRIX_VARIABLES_NUMBER:
                return True
        return False

    @staticmethod
    def _internal_render(widget, data, **kwargs):
        fig = widget(data, fontsize=10, **kwargs)
        plt.show(fig)

    def _get_operation(self, graph_type):
        return self.__OPERATIONS_MAPPING[graph_type]
